import numpy as np


class Constants(object):
    '''Class holding global constants'''

    RAD2ARCSEC = 180 * 3600 / np.pi
    ARCSEC2RAD = np.pi / (180 * 3600)
    DEG2RAD = np.pi / 180


RAD2ARCSEC = 180 * 3600 / np.pi
ARCSEC2RAD = np.pi / (180 * 3600)
DEG2RAD = np.pi / 180
